function XLuaClass(className, super, argList)
    local child = {}
    if argList then
        for key, value in pairs(argList) do
            child[key] = value
        end
    end
    child.XClassName = className
    setmetatable(child, {__index = super})
    return child
end

function xe:Class(...)
    return XLuaClass("UnnamedClass", ...)
end

function xe.XClass(...)
    return XLuaClass(...)
end

local function emptyFunc()
end

local engineIns = _G_EngineInstance

local _XEWorldController = {}

local _XEDirector = {}
_XEDirector.XType = "xes::Director"
_XEDirector.__wcArr__ = {}

function _XEDirector:GetInstance()
    return _XEDirector
end

function _XEDirector:GetDressupManager()
    return engineIns:GetDressupManager()
end

function _XEDirector:GetViewport()
    return engineIns:GetCurViewPort()
end

function _XEDirector:GetEventDispatcher()
    return engineIns:GetEventDispatcher()
end

function _XEDirector:SetEngineLogEnable(b)
    engineIns:SetLogEnable(b)
end

function _XEDirector:SetLogEnable(b)
    engineIns:SetLogEnable(b)
end

function _XEDirector:GetAnimationManager()
    return engineIns:GetAnimationManager()
end

function _XEDirector:GetRunningWorldController()
    return self.__wcArr__
end

function _XEDirector:GetVersion()
    return engineIns:GetEngineVersion()
end

function _XEDirector:GetVersionNum()
    return engineIns:GetEngineVersionValue()
end

function _XEDirector:SetTag(tag)
    engineIns:SetTag(tag)
end

function _XEDirector:GetTag()
    return engineIns:GetTag()
end

function _XEDirector:GetPlatform()
    return XEDevice:GetPlatform()
end

function _XEDirector:ShowDebugInfo(b)
    XEUtility.ShowDebugInfo(b)
end

function _XEDirector:GetWindow()
    return engineIns:GetPlatformWindow()
end

function _XEDirector:GetEngine()
    return engineIns
end

function _XEDirector:IsRunning()
    return true
end

function _XEDirector:PushWorldController(worldController)
    if not self.__tickListener__ then
        self.__tickListener__ = XEDataEventListener:Create()
        self.__tickListener__:RegisterHandler(
            function()
                local wcnum = #self.__wcArr__
                if wcnum > 0 then
                    local wc = self.__wcArr__[wcnum]
                    wc:__tick__()
                end
            end,
            "onTick"
        )
        engineIns:GetEventDispatcher():AddEventListener(self.__tickListener__, nil)
    end

    table.insert(self.__wcArr__, worldController)
    local world = worldController:GetWorld()
    engineIns:GetCurViewPort():AttachWorld(world)
    if worldController.onStart then
        worldController:onStart()
    end
    local preWorldController = self.__wcArr__[#self.__wcArr__ - 1]
    if preWorldController and preWorldController.onPause then
        preWorldController:onPause()
    end
    if worldController.onResume then
        worldController:onResume()
    end
end

function _XEDirector:PopWorldController()
    local wcNum = #self.__wcArr__

    local removeWorldController = self.__wcArr__[wcNum]

    if removeWorldController then
        if removeWorldController.onPause then
            removeWorldController:onPause()
        end

        local world = removeWorldController:GetWorld()
        engineIns:DeleteXEWorld(world)

        table.remove(self.__wcArr__, wcNum)

        local resumeWorldController = self.__wcArr__[#self.__wcArr__]
        if resumeWorldController then
            world = resumeWorldController:GetWorld()
            engineIns:GetCurViewPort():AttachWorld(world)

            if resumeWorldController.onResume then
                resumeWorldController:onResume()
            end
        end

        engineIns:GetEventDispatcher():RemoveEventListenersForTarget(removeWorldController)
        if removeWorldController.onEnd then
            removeWorldController:onEnd()
        end
    end
end

function _XEDirector:GetTopWorldController()
    return self.__wcArr__[#self.__wcArr__]
end

function _XEDirector:ClearBackground()
    engineIns:GetCurViewPort():GetXViewport():SetClearColor(0)
end

function _XEDirector:EnableClearColor(b)
    engineIns:GetCurViewPort():GetXViewport():EnableClearColor(b)
end

function _XEDirector:GetBasePath()
    return engineIns:GetBaseDir()
end
function _XEDirector:GetLibraryPath()
    return engineIns:GetLibraryDir()
end

function _XEDirector.GetXUIManager()
    return engineIns:GetUIManager()
end

function _XEDirector.GetXCanvasBase()
    return engineIns:GetCanvasBase()
end

function _XEDirector.GetXWindow()
    return _XEDirector:GetWindow()
end

function _XEDirector.GetX2DPhysicalManager()
    return engineIns:Get2DPhysicalManager()
end

function _XEDirector.GetXPhysicalManager()
    return engineIns:GetPhysicalManager()
end

function _XEDirector.GetXModelManager()
    return engineIns:GetModelManager()
end

function _XEDirector.GetXEditorModelManager()
    return engineIns:GetEditorModelManager()
end

function _XEDirector.GetXGestureManager()
    return engineIns:GetGestureManager()
end

_XEDirector.PushScene = _XEDirector.PushWorldController
_XEDirector.PopScene = _XEDirector.PopWorldController
_XEDirector.GetTopScene = _XEDirector.GetTopWorldController
_XEDirector.PresentScene = _XEDirector.PresentWorldController
_XEDirector.DissmissScene = _XEDirector.DissmissWorldController
_XEDirector.GetPresentScene = _XEDirector.GetPresentWorldController
_XEDirector.GetRunningScene = _XEDirector.GetRunningWorldController
_XEDirector.GetEngineVersion = _XEDirector.GetVersion
_XEDirector.GetEngineVersionNum = _XEDirector.GetVersionNum

function _XEWorldController:new(name)
    local instance = {
        __name__ = name,
        __world__ = engineIns:CreateXEWorld()
    }
    setmetatable(instance, {__index = self})
    return instance
end

function _XEWorldController:GetWorld()
    return self.__world__
end

function _XEWorldController:Create2DPhysicalScene(name)
    self.__physicalScene2D__ = engineIns:Get2DPhysicalManager():CreateScene(name)
    return self.__physicalScene2D__
end

function _XEWorldController:__tick__()
    local fDelta = engineIns:GetCurDeltaTime()
    if self.__physicalScene2D__ then
        self.__physicalScene2D__:Tick(fDelta)
    end
    if self.onTick then
        self:onTick(fDelta)
    end
end

_XEWorldController.Create = _XEWorldController.new
_XEWorldController.CreateScene = _XEWorldController.new
_XEWorldController.StartTick = emptyFunc
_XEWorldController.startTick = emptyFunc
_XEWorldController.StopTick = emptyFunc
_XEWorldController.stopTick = emptyFunc

if IX2DPhysicalManager then
    function IX2DPhysicalManager:Init()
        print_warning("IX2DPhysicalManager:Init 无效API")
    end
    function IX2DPhysicalManager:Release()
        print_warning("IX2DPhysicalManager:Release 无效API")
    end
    function IX2DPhysicalManager:GetGlobal()
        return engineIns:Get2DPhysicalManager()
    end
end

function IXPlatformWindow:GetGlobal()
    return _XEDirector:GetWindow()
end

g_pXWindow = engineIns:GetPlatformWindow()
g_pIX2DPhysicalManager = engineIns:Get2DPhysicalManager()
g_pIXPhysicalManager = engineIns:GetPhysicalManager()
g_pIXGestureManager = engineIns:GetGestureManager()
g_pFxManager = engineIns:GetFxManager()
g_pXAnimationManager = engineIns:GetAnimationManager()
g_pBrushManager = engineIns:GetBrushManager()
g_pXCanvasBase = engineIns:GetCanvasBase()
g_pXConVarSys = engineIns:GetConVarSys()
g_pDressupManager = engineIns:GetDressupManager()
g_pIXGestureManager = engineIns:GetGestureManager()
g_pXLightMgr = engineIns:GetLightManager()
g_pXMaterialMgr = engineIns:GetMaterialManager()
g_pXUserMaterialMgr = engineIns:GetUserMaterialManager()
g_pXModelManager = engineIns:GetModelManager()
g_pXCurrentRenderer = engineIns:GetRenderer()
g_pXTextureManager = engineIns:GetTextureManager()

function XEEventDispatcher:GetInstance()
    return engineIns:GetEventDispatcher()
end

XECast = tolua.cast
XEWorldController = _XEWorldController
xe.GestureEventListener = XEGestureEventListener
xe.GestureEventListenerV1 = XEGestureEventListener
xe.Scene = _XEWorldController
xe.Device = XEDevice
xe.Scheduler = XEScheduler
xe.EventDispatcher = XEEventDispatcher
xe.DataEventListener = XEDataEventListener
xe.Director = _XEDirector
_G_Director = _XEDirector

xe.App = xe.App or {}
xe.App.Platform = _XEDirector:GetPlatform()
xe.PLATFORM_OS_NONE = 0
xe.PLATFORM_OS_IOS = 1
xe.PLATFORM_OS_ANDROID = 2

function FileUtils:FilePath(path)
    return path
end
