math.mod = math.mod or math.fmod

function math.isNumber(v)
    return type(v) == "number"
end
function math.isVec2(v)
    return type(v) == "table" and v.XType == "XVECTOR2"
end
function math.isVec3(v)
    return type(v) == "table" and v.XType == "XVECTOR3"
end
function math.isVec4(v)
    return type(v) == "table" and v.XType == "XVECTOR4"
end
function math.isM3(v)
    return type(v) == "table" and v.XType == "XMATRIX3"
end
function math.isM4(v)
    return type(v) == "table" and v.XType == "XMATRIX4"
end
function math.isQuaternion(v)
    return type(v) == "table" and v.XType == "XQUATERNION"
end

function math.typeName(v)
    if v.XType then
        return v.XType
    end
    return type(v)
end

function math.mul_m3_vec2(mat, v)
    local w = v.x * mat._13 + v.y * mat._23 + mat._33
    if (math.abs(w) < 1e-6) then
        return XVECTOR2(0)
    end

    local vRet = XVECTOR2(v.x * mat._11 + v.y * mat._21 + mat._31, v.x * mat._12 + v.y * mat._22 + mat._32)
    return vRet / w
end
function math.mul_m3_vec3(mat, v)
    return XVECTOR3(
        v.x * mat._11 + v.y * mat._21 + v.z * mat._31,
        v.x * mat._12 + v.y * mat._22 + v.z * mat._32,
        v.x * mat._13 + v.y * mat._23 + v.z * mat._33
    )
end

function math.mul_m3_m3(mat1, mat2)
    return XMATRIX3(
        mat1._11 * mat2._11 + mat1._12 * mat2._21 + mat1._13 * mat2._31,
        mat1._11 * mat2._12 + mat1._12 * mat2._22 + mat1._13 * mat2._32,
        mat1._11 * mat2._13 + mat1._12 * mat2._23 + mat1._13 * mat2._33,

        mat1._21 * mat2._21 + mat1._22 * mat2._21 + mat1._23 * mat2._31,
        mat1._21 * mat2._22 + mat1._22 * mat2._22 + mat1._23 * mat2._32,
        mat1._21 * mat2._23 + mat1._22 * mat2._23 + mat1._23 * mat2._33,

        mat1._31 * mat2._21 + mat1._32 * mat2._21 + mat1._33 * mat2._31,
        mat1._31 * mat2._22 + mat1._32 * mat2._22 + mat1._33 * mat2._32,
        mat1._31 * mat2._23 + mat1._32 * mat2._23 + mat1._33 * mat2._33
    )
end

function math.mul_m4_float(mat, float)
    return XMATRIX4(
        mat._11 * float,
        mat._12 * float,
        mat._13 * float,
        mat._14 * float,
        mat._21 * float,
        mat._22 * float,
        mat._23 * float,
        mat._24 * float,
        mat._31 * float,
        mat._32 * float,
        mat._33 * float,
        mat._34 * float,
        mat._41 * float,
        mat._42 * float,
        mat._43 * float,
        mat._44 * float
    )
end
function math.mul_m4_vec3(mat, vec3)
    local vRet =
        XVECTOR3(
        vec3.x * mat._11 + vec3.y * mat._21 + vec3.z * mat._31 + mat._41,
        vec3.x * mat._12 + vec3.y * mat._22 + vec3.z * mat._32 + mat._42,
        vec3.x * mat._13 + vec3.y * mat._23 + vec3.z * mat._33 + mat._43
    )
    local ww = (vec3.x * mat._14 + vec3.y * mat._24 + vec3.z * mat._34 + mat._44)
    if ww == 0 then
        error("错误计算 mul_m4_vec3")
    end
    local w = 1 / (vec3.x * mat._14 + vec3.y * mat._24 + vec3.z * mat._34 + mat._44)

    return vRet:MulFloatWith(w)
end
function math.mul_m4_vec4(mat, vec4)
    return XVECTOR4(
        vec4.x * mat._11 + vec4.y * mat._21 + vec4.z * mat._31 + vec4.w * mat._41,
        vec4.x * mat._12 + vec4.y * mat._22 + vec4.z * mat._32 + vec4.w * mat._42,
        vec4.x * mat._13 + vec4.y * mat._23 + vec4.z * mat._33 + vec4.w * mat._43,
        vec4.x * mat._14 + vec4.y * mat._24 + vec4.z * mat._34 + vec4.w * mat._44
    )
end
function math.mul_m4_m4(mat1, mat2)
    return XMATRIX4(
        mat1._11 * mat2._11 + mat1._12 * mat2._21 + mat1._13 * mat2._31 + mat1._14 * mat2._41,
        mat1._11 * mat2._12 + mat1._12 * mat2._22 + mat1._13 * mat2._32 + mat1._14 * mat2._42,
        mat1._11 * mat2._13 + mat1._12 * mat2._23 + mat1._13 * mat2._33 + mat1._14 * mat2._43,
        mat1._11 * mat2._14 + mat1._12 * mat2._24 + mat1._13 * mat2._34 + mat1._14 * mat2._44,

        mat1._21 * mat2._21 + mat1._22 * mat2._21 + mat1._23 * mat2._31 + mat1._24 * mat2._41,
        mat1._21 * mat2._22 + mat1._22 * mat2._22 + mat1._23 * mat2._32 + mat1._24 * mat2._42,
        mat1._21 * mat2._23 + mat1._22 * mat2._23 + mat1._23 * mat2._33 + mat1._24 * mat2._43,
        mat1._21 * mat2._24 + mat1._22 * mat2._24 + mat1._23 * mat2._34 + mat1._24 * mat2._44,

        mat1._31 * mat2._21 + mat1._32 * mat2._21 + mat1._33 * mat2._31 + mat1._34 * mat2._41,
        mat1._31 * mat2._22 + mat1._32 * mat2._22 + mat1._33 * mat2._32 + mat1._34 * mat2._42,
        mat1._31 * mat2._23 + mat1._32 * mat2._23 + mat1._33 * mat2._33 + mat1._34 * mat2._43,
        mat1._31 * mat2._24 + mat1._32 * mat2._24 + mat1._33 * mat2._34 + mat1._34 * mat2._44,

        mat1._41 * mat2._21 + mat1._42 * mat2._21 + mat1._43 * mat2._31 + mat1._44 * mat2._41,
        mat1._41 * mat2._22 + mat1._42 * mat2._22 + mat1._43 * mat2._32 + mat1._44 * mat2._42,
        mat1._41 * mat2._23 + mat1._42 * mat2._23 + mat1._43 * mat2._33 + mat1._44 * mat2._43,
        mat1._41 * mat2._24 + mat1._42 * mat2._24 + mat1._43 * mat2._34 + mat1._44 * mat2._44
    )
end
function math.mul_quaternion_vec3(q, v)
    local vq = XQUATERNION(1, v)
    local qi = XQUATERNION(q.w, -q.x, -q.y, -q.z)
    local qvq = q * vq * qi

    return XVECTOR3(qvq.x, qvq.y, qvq.z)
end

function X_Equals(a, b, esp)
    return math.abs(a - b) < esp
end
function LERPVec(vec1, vec2, fraction)
    return vec1 * (1.0 - fraction) + vec2 * fraction
end
function CrossProduct(arg1, arg2)
    if math.isVec2(arg1) and math.isVec2(arg2) then
        return arg1.y * arg2.y - arg1.y * arg2.x
    elseif math.isVec3(arg1) and math.isVec3(arg2) then
        return XVECTOR3(
            arg1.y * arg2.z - arg1.z * arg2.y,
            arg1.z * arg2.x - arg1.x * arg2.z,
            arg1.x * arg2.y - arg1.y * arg2.x
        )
    end
end
function DotProduct(arg1, arg2)
    if math.isVec3(arg1) and math.isVec3(arg2) then
        return arg1.x * arg2.x + arg1.y * arg2.y + arg1.z * arg2.z
    elseif math.isVec2(arg1) and math.isVec2(arg2) then
        return arg1.x * arg2.x + arg1.y * arg2.y
    elseif math.isQuaternion(arg1) and math.isQuaternion(arg2) then
        return arg1.w * arg2.w + arg1.x * arg2.x + arg1.y * arg2.y + arg1.z * arg2.z
    elseif math.isNumber(arg1) and math.isNumber(arg2) then
        return arg1 + arg2
    end
end
function Normalize(arg)
    if math.isVec3(arg) then
        local retVec = XVECTOR3(arg)
        local ret = retVec:Normalize()
        return ret, retVec
    elseif math.isQuaternion(arg) then
        local retQ = XQUATERNION(arg)
        return retQ:Normalize()
    end
end

__XG_CastRay__ = XG_CastRay
function XG_CastRay(...)
    local args = {...}
    if #args == 2 then
        return __XG_CastRay__(args[1], args[2])
    elseif #args == 4 then
        print("您应该使用 local start, end_ = XG_CastRay(viewport, vec3) 的方式以提高性能")
        local start, end_ = __XG_CastRay__(args[1], args[2])
        args[3]:Set(start.x, start.y, start.z)
        args[4]:Set(end_.x, end_.y, end_.z)
    end
end
