XMATRIX3 = { XType = 'XMATRIX3', IDENTITY = 0, CLEARED = 1 }
setmetatable(XMATRIX3, XMATRIX3)

function XMATRIX3:GetRow(i)
    local m = self.m
    return XVECTOR3(m[i][1], m[i][2], m[i][3])
end
function XMATRIX3:GetCol(i)
    local m = self.m
    return XVECTOR3(m[1][i], m[2][i], m[3][i])
end
function XMATRIX3:Transpose()
    local t = self._12
    self._12 = self._21
    self._21 = t

    t = self._13
    self._13 = self._31
    self._31 = t

    t = self._23
    self._23 = self._32
    self._32 = t
end
function XMATRIX3:Clear()
    self._11 = 0
    self._12 = 0
    self._13 = 0
    self._21 = 0
    self._22 = 0
    self._23 = 0
    self._31 = 0
    self._32 = 0
    self._33 = 0
end
function XMATRIX3:Identity()
    self._11 = 1
    self._12 = 0
    self._13 = 0
    self._21 = 0
    self._22 = 1
    self._23 = 0
    self._31 = 0
    self._32 = 0
    self._33 = 1
end
function XMATRIX3:Translate(x, y)
    self:Identity()
    self._31 = x
    self._32 = y
end
function XMATRIX3:Rotate(fRad)
    self:Identity()
    self._11 = math.cos(fRad)
    self._22 = self._11
    self._12 = math.sin(fRad)
    self._21 = -self._12
end
function XMATRIX3:InverseTM()
    local m = self.m
    local matInverse = { {}, {}, {} }

    matInverse[1][1] = m[2][2] * m[3][3] - m[2][3] * m[3][2]
    matInverse[1][2] = m[1][3] * m[3][2] - m[1][2] * m[3][3]
    matInverse[1][3] = m[1][2] * m[2][3] - m[1][3] * m[2][2]
    matInverse[2][1] = m[2][3] * m[3][1] - m[2][1] * m[3][3]
    matInverse[2][2] = m[1][1] * m[3][3] - m[1][3] * m[3][1]
    matInverse[2][3] = m[1][3] * m[2][1] - m[1][1] * m[2][3]
    matInverse[3][1] = m[2][1] * m[3][2] - m[2][2] * m[3][1]
    matInverse[3][2] = m[1][2] * m[3][1] - m[1][1] * m[3][2]
    matInverse[3][3] = m[1][1] * m[2][2] - m[1][2] * m[2][1]

    local fDet = m[1][1] * matInverse[1][1] +
            m[1][2] * matInverse[2][1] +
            m[1][3] * matInverse[3][1]

    local fInvDet = 1 / fDet

    for iRow = 1, 3 do
        for iCol = 1, 3 do
            m[iRow][iCol] = matInverse[iRow][iCol] * fInvDet
        end
    end
    self._11 = m[1][1]
    self._12 = m[1][2]
    self._13 = m[1][3]
    self._21 = m[2][1]
    self._22 = m[2][2]
    self._23 = m[2][3]
    self._31 = m[3][1]
    self._32 = m[3][2]
    self._33 = m[3][3]

end

function XMATRIX3.__mul(arg1, arg2)

    if math.isVec2(arg1) and math.isM3(arg2) then
        --- vec2 * m3
        return math.mul_m3_vec2(arg2, arg1)
    elseif math.isVec2(arg2) and math.isM3(arg1) then
        --- m3 * vec2
        return math.mul_m3_vec2(arg1, arg2)
    elseif math.isVec3(arg1) and math.isM3(arg2) then
        --- vec3 * m3
        return math.mul_m3_vec3(arg2, arg1)
    elseif math.isVec3(arg2) and math.isM3(arg1) then
        --- m3 * vec3
        return math.mul_m3_vec3(arg1, arg2)
    elseif math.isM3(arg1) and math.isM3(arg2) then
        --- m3 * m3
        return math.mul_m3_m3(arg1, arg2)
    end

end
function XMATRIX3.__eq(arg1, arg2)
    if math.isM3(arg1) and math.isM3(arg2) then
        return arg1._11 == arg2._11 and arg1._12 == arg2._12 and arg1._13 == arg2._13 and
                arg1._21 == arg2._21 and arg1._22 == arg2._22 and arg1._23 == arg2._23 and
                arg1._31 == arg2._31 and arg1._32 == arg2._32 and arg1._33 == arg2._33
    end
    return false
end
function XMATRIX3:__index(key)
    if key == 'm' then
        return { { self._11, self._12, self._13 },
                 { self._21, self._22, self._23 },
                 { self._31, self._32, self._33 } }
    end
    return rawget(XMATRIX3, key)
end
function XMATRIX3:__tostring()
    return '\nXMATRIX3:[ ' .. tostring(self._11) .. ' ' .. tostring(self._12) .. ' ' .. tostring(self._13) .. ' ' .. '\n' ..
            '           ' .. tostring(self._21) .. ' ' .. tostring(self._22) .. ' ' .. tostring(self._23) .. ' ' .. '\n' ..
            '           ' .. tostring(self._31) .. ' ' .. tostring(self._32) .. ' ' .. tostring(self._33) .. ' ' .. ' ]'

end
function XMATRIX3:__call(...)
    local m3 = {}
    setmetatable(m3, XMATRIX3)

    local args = { ... }
    if #args == 0 then
        self:Clear()
    elseif #args == 1 then
        local arg = args[1]
        if math.isNumber(arg) and arg == XMATRIX3.CLEARED then
            self:Clear()
        elseif math.isNumber(arg) and arg == XMATRIX3.IDENTITY then
            self:Identity()
        elseif math.isM3(arg) then

            m3._11 = arg._11
            m3._12 = arg._12
            m3._13 = arg._13
            m3._21 = arg._21
            m3._22 = arg._22
            m3._23 = arg._23
            m3._31 = arg._31
            m3._32 = arg._32
            m3._33 = arg._33

        elseif type(arg) == 'table' and #arg == 9 then
            m3._11 = arg[1]
            m3._12 = arg[2]
            m3._13 = arg[3]
            m3._21 = arg[4]
            m3._22 = arg[5]
            m3._23 = arg[6]
            m3._31 = arg[7]
            m3._32 = arg[8]
            m3._33 = arg[9]

        end
    elseif #args == 9 then
        m3._11 = args[1]
        m3._12 = args[2]
        m3._13 = args[3]
        m3._21 = args[4]
        m3._22 = args[5]
        m3._23 = args[6]
        m3._31 = args[7]
        m3._32 = args[8]
        m3._33 = args[9]
    end

    return m3

end
