XQUATERNION = { XType = 'XQUATERNION' }
setmetatable(XQUATERNION, XQUATERNION)

function XQUATERNION:ConvToMatrix()
    return QuaternionToMatrix(self)
end
function XQUATERNION:ConvFromMatrix(mat)
    local q = MatrixToQuaternion(mat)
    self.x = q.x
    self.y = q.y
    self.z = q.z
    self.w = q.w
end
function XQUATERNION:ConvToEulerAngle()
    local _, x, y, z = QuaternionToEulerAngle(self)
    return x, y, z
end
function XQUATERNION:ConvFromEulerAngle(fXAngle, fYAngle, fZAngle)
    self.x, self.y, self.z, self.w = EulerAngleToQuaternionValue(fXAngle, fYAngle, fZAngle)
    return self
end
function XQUATERNION:ConvertToAxisAngle()
    return QuaternionToAxisAngle(self)
end
function XQUATERNION:ConvertFromAxisAngle(vAxis, fAngle)
    local q = AxisAngleToQuaternion(vAxis, fAngle)
    self:SetWithQuaternion(q)
end
function XQUATERNION:RotationFromTo()
    local q = QuaternionRotationFromTo(vAxis, fAngle)
    self:SetWithQuaternion(q)
    return self
end
function XQUATERNION:Clear()
    self.x = 0
    self.y = 0
    self.z = 0
    self.w = 0
    return self
end
function XQUATERNION:Identity()
    self.x = 0
    self.y = 0
    self.z = 0
    self.w = 1
    return self
end
function XQUATERNION:Set(x, y, z, w)
    self.x = x
    self.y = y
    self.z = z
    self.w = w
    return self
end
function XQUATERNION:SetWithQuaternion(q)
    self.x = q.x
    self.y = q.y
    self.z = q.z
    self.w = q.w
end
function XQUATERNION:Norm()
    return math.sqrt(self.x ^ 2 + self.y ^ 2 + self.z ^ 2 + self.w ^ 2)
end
function XQUATERNION:Normalize()
    local f = 1 / self:Norm()
    self.x = self.x * f
    self.y = self.y * f
    self.z = self.z * f
    self.w = self.w * f
    return self
end
function XQUATERNION:Conjugate()
    self.x = -self.x
    self.y = -self.y
    self.z = -self.z
    return self
end
function XQUATERNION:Inverse()
    local f = 1 / self:Norm()
    self.x = self.x * f
    self.y = self.y * -f
    self.z = self.z * -f
    self.w = self.w * -f
end

function XQUATERNION.__add(q1, q2)
    if math.isQuaternion(q1) and math.isQuaternion(q2) then
        return XQUATERNION(q1.w + q2.w, q1.x + q2.x, q1.y + q2.y, q1.z + q2.z)
    end
    error('Illegal operation: ' .. math.typeName(q1) .. ' + ' .. math.typeName(q2))
end
function XQUATERNION.__sub(q1, q2)
    if math.isQuaternion(q1) and math.isQuaternion(q2) then
        return XQUATERNION(q1.w - q2.w, q1.x - q2.x, q1.y - q2.y, q1.z - q2.z)
    end
    error('Illegal operation: ' .. math.typeName(q1) .. ' - ' .. math.typeName(q2))
end
function XQUATERNION.__mul(arg1, arg2)
    if math.isQuaternion(arg1) and math.isNumber(arg2) then
        return XQUATERNION(arg1.w * arg2, arg1.x * arg2, arg1.y * arg2, arg1.z * arg2)
    elseif math.isQuaternion(arg2) and math.isNumber(arg1) then
        return XQUATERNION(arg2.w * arg1, arg2.x * arg1, arg2.y * arg1, arg2.z * arg1)
    elseif math.isQuaternion(arg1) and math.isVec3(arg2) then
        return math.mul_quaternion_vec3(arg1, arg2)
    elseif math.isQuaternion(arg1) and math.isQuaternion(arg2) then
        return XQUATERNION(arg1.w * arg2.w - arg1.x * arg2.x - arg1.y * arg2.y - arg1.z * arg2.z,
                arg1.w * arg2.x + arg1.x * arg2.w + arg1.y * arg2.z - arg1.z * arg2.y,
                arg1.w * arg2.y + arg1.y * arg2.w + arg1.z * arg2.x - arg1.x * arg2.z,
                arg1.w * arg2.z + arg1.z * arg2.w + arg1.x * arg2.y - arg1.y * arg2.x)
    end
    error('Illegal operation: ' .. math.typeName(arg1) .. ' * ' .. math.typeName(arg2))
end
function XQUATERNION.__div(arg1, arg2)

    if math.isQuaternion(arg1) and math.isNumber(arg2) then
        local fValue = 1 / arg2
        return XQUATERNION(arg1.w * fValue, arg1.x * fValue, arg1.y * fValue, arg1.z * fValue)
    end
    error('Illegal operation: ' .. math.typeName(arg1) .. ' / ' .. math.typeName(arg2))
end
function XQUATERNION.__pow(q, v)
    if math.isQuaternion(q) and math.isVec3(v) then

        local uv = XVECTOR3()
        local uuv = XVECTOR3()
        local qvec = XVECTOR3(q.x, q.y, q.z)

        uv = CrossProduct(qvec, v)
        uuv = CrossProduct(qvec, uv)
        uv = (2 * q.w) * uv
        uuv = 2 * uuv

        return v + uv + uuv
    end

    error('Illegal operation: ' .. math.typeName(q) .. ' ^ ' .. math.typeName(v))
end
function XQUATERNION.__eq(arg1, arg2)
    if math.isQuaternion(arg2) and math.isQuaternion(arg1) then
        return arg1.x == arg2.x and arg1.y == arg2.y and arg1.z == arg2.z and arg1.w == arg2.w
    end
    return false
end
function XQUATERNION:__tostring()
    return 'XQUATERNION(x=' .. self.x .. ', y=' .. self.y .. ', z=' .. self.z .. ', w=' .. self.w .. ')'
end
function XQUATERNION:__index(key)
    if key == 'm' then
        return { self.x, self.y, self.z, self.w }
    end
    return rawget(XQUATERNION, key)
end
function XQUATERNION:__call(...)
    local val = {}
    setmetatable(val, XQUATERNION)
    local args = { ... }
    if #args == 0 then
        val.x = 0
        val.y = 0
        val.z = 0
        val.w = 0
    elseif #args == 1 then
        local arg = args[1]
        if math.isQuaternion(arg) then
            self:SetWithQuaternion(arg)
        else
            error('XQUATERNION() 错误的参数类型 ' .. math.typeName(arg))
        end
    elseif #args == 2 then
        if math.isVec3(args[1]) and math.isNumber(args[2]) then
            local q = AxisAngleToQuaternion(args[1], args[2])
            self:SetWithQuaternion(q)
        elseif math.isVec3(args[2]) and math.isNumber(args[1]) then
            local q = AxisAngleToQuaternion(args[2], args[1])
            self:SetWithQuaternion(q)
        end

    elseif #args == 3 then
        --- 欧拉角到四元数
        local q = EulerAngleToQuaternion(args[1], args[2], args[3])
        self.x = q.x
        self.y = q.y
        self.z = q.z
        self.w = q.w

    elseif #args == 4 then
        val.w = args[1]
        val.x = args[2]
        val.y = args[3]
        val.z = args[4]
    else
        error('XQUATERNION() 错误的参数个数 ' .. #args)
    end

    return val
end
