XVECTOR4 = { XType = 'XVECTOR4' }
setmetatable(XVECTOR4, XVECTOR4)

function XVECTOR4:Set(x, y, z, w)
    self.x = x
    self.y = y
    self.z = z
    self.w = w
end
function XVECTOR4:Clear()
    self.x = 0
    self.y = 0
    self.z = 0
    self.w = 0
end
function XVECTOR4:IsZero()
    return self.x == 0 and self.y == 0 and self.z == 0 and self.w == 0
end
function XVECTOR4:IsEquals(vec, epsilon)
    return X_Equals(self.x, vec.x, epsilon) and X_Equals(self.y, vec.y, epsilon) and X_Equals(self.z, vec.z, epsilon) and X_Equals(self.w, vec.w, epsilon)
end
function XVECTOR4:MinMember()
    return math.min(self.x, self.y, self.z, self.w)
end
function XVECTOR4:MaxMember()
    return math.max(self.x, self.y, self.z, self.w)
end

function XVECTOR4.__add(arg1, arg2)
    if math.isVec4(arg2) then
        return XVECTOR4(arg1.x + arg2.x, arg1.y + arg2.y, arg1.z + arg2.z, arg1.w + arg2.w)
    end
    error('Illegal operation: ' .. math.typeName(arg1) .. ' + ' .. math.typeName(arg2))
end
function XVECTOR4.__sub(arg1, arg2)
    if math.isVec4(arg2) then
        return XVECTOR4(arg1.x - arg2.x, arg1.y - arg2.y, arg1.z - arg2.z, arg1.w - arg2.w)
    end
    error('Illegal operation: ' .. math.typeName(arg1) .. ' - ' .. math.typeName(arg2))
end
function XVECTOR4.__mul(arg1, arg2)
    if math.isVec4(arg2) and type(arg1) == 'number' then
        return XVECTOR4(arg1 * arg2.x, arg1 * arg2.y, arg1 * arg2.z, arg1 * arg2.w)
    elseif math.isVec4(arg1) and type(arg2) == 'number' then
        return XVECTOR4(arg1.x * arg2, arg1.y * arg2, arg1.z * arg2, arg1.w * arg2)
    elseif math.isVec4(arg1) and math.isVec4(arg2) then
        return XVECTOR4(arg1.x * arg2.x, arg1.y * arg2.y, arg1.z * arg2.z, arg1.w * arg2.w)
    end
    error('Illegal operation: ' .. math.typeName(arg1) .. ' * ' .. math.typeName(arg2))
end
function XVECTOR4.__div(arg1, arg2)
    if type(arg2) == 'number' then
        return XVECTOR4(arg1.x / arg2, arg1.y / arg2, arg1.z / arg2, arg1.w / arg2)
    elseif math.isVec4(arg2) then
        return XVECTOR4(arg1.x / arg2.x, arg1.y / arg2.y, arg1.z / arg2.z, arg1.w / arg2)
    end
    error('Illegal operation: ' .. math.typeName(arg1) .. ' / ' .. math.typeName(arg2))
end
function XVECTOR4.__eq(arg1, arg2)
    if math.isVec4(arg1) and math.isVec4(arg2) then
        return arg1.x == arg2.x and arg1.y == arg2.y and arg1.z == arg2.z and arg1.w == arg2.w
    end
    return false
end
function XVECTOR4:__tostring()
    return 'XVECTOR4(x=' .. self.x .. ', y=' .. self.y .. ', z=' .. self.z .. ', w=' .. self.w .. ')'
end
function XVECTOR4:__index(key)
    if key == 'm' then
        return { self.x, self.y, self.z, self.w }
    end
    return rawget(XVECTOR4, key)
end
function XVECTOR4:__call(...)
    local vec4 = {}
    local args = { ... }
    if #args == 0 then
        vec4.x = 0
        vec4.y = 0
        vec4.z = 0
        vec4.w = 0
    elseif #args == 1 then
        local arg = args[1]
        if type(arg) == 'number' then
            vec4.x = arg
            vec4.y = arg
            vec4.z = arg
            vec4.w = arg
        elseif type(arg) == 'table' and math.isVec3(arg) then
            vec4.x = arg.x
            vec4.y = arg.y
            vec4.z = arg.z
            vec4.w = 1
        elseif type(arg) == 'table' and math.isVec4(arg) then
            vec4.x = arg.x
            vec4.y = arg.y
            vec4.z = arg.z
            vec4.w = arg.w
        else
            error('XVECTOR4() 错误的参数类型 ' .. type(arg))
        end
    elseif #args == 4 then
        vec4.x = args[1]
        vec4.y = args[2]
        vec4.z = args[3]
        vec4.w = args[4]
    else
        error('XVECTOR4() 错误的参数个数 ' .. #args)
    end
    setmetatable(vec4, XVECTOR4)
    return vec4
end
