function table.append(a, b)
    for _, v in ipairs(b) do
        a[#a + 1] = v
    end
end
function table.clear(t)
    for k in pairs(t) do
        t[k] = nil
    end
end
function table.count(t, value)
    local count = 0
    for _, v in pairs(t) do
        if (typeof(v, "table") and typeof(value, "table")) then
            if (table.equal(v, value)) then
                count = count + 1
            end
        elseif (v == value) then
            count = count + 1
        end
    end
    return count
end
function table.equal(a, b)
    if (getmetatable(a) ~= getmetatable(b)) then
        return false
    end

    for k, v in pairs(a) do
        if (typeof(v, "table") and
                typeof(b[k], "table") and
                not table.equal(v, b[k])) then
            return false
        elseif (v ~= b[k]) then
            return false
        end
    end

    for k, v in pairs(b) do
        if (typeof(v, "table") and
                typeof(a[k], "table") and
                not table.equal(v, a[k])) then
            return false
        elseif (v ~= a[k]) then
            return false
        end
    end
    return true
end
function table.hasvalue(t, value)
    for k, v in pairs(t) do
        if (v == value) then
            return k
        end
    end
    return nil
end
function table.len(t)
    local c = 0
    for _ in pairs(t) do
        c = c + 1
    end
    return c
end
function table.merge(a, b)
    for k, v in pairs(b) do
        a[k] = v
    end
end
function table.prepose(a, b)
    for i = #b, 1, -1 do
        table.insert(a, 1, b[i])
    end
end
function table.print(t, i, printed)
    if (t == nil) then
        print(nil)
        return
    end

    printed = printed or {}
    printed[t] = true

    i = i or 0
    local indent = string.rep("\t", i)
    for k, v in pairs(t) do
        if (type(v) == "table" and not printed[v]) then
            print(indent .. k)
            table.print(v, i + 1, printed)
        else
            print(indent .. tostring(k), v)
        end
    end
end
function table.removevalue(t, v)
    for i, w in ipairs(t) do
        if (w == v) then
            table.remove(t, i)
        end
    end
end
function table.unique(t)
    local copy = {}
    for _, v in ipairs(t) do
        if (table.count(copy, v) == 0) then
            copy[#copy + 1] = v
        end
    end
    return copy
end
function table.nums(t)
    local count = 0
    for k, v in pairs(t) do
        count = count + 1
    end
    return count
end
function table.keys(hashtable)
    local keys = {}
    for k, v in pairs(hashtable) do
        keys[#keys + 1] = k
    end
    return keys
end
function table.values(hashtable)
    local values = {}
    for k, v in pairs(hashtable) do
        values[#values + 1] = v
    end
    return values
end
function table.merge(dest, src)
    for k, v in pairs(src) do
        dest[k] = v
    end
end
function table.insertto(dest, src, begin)
    begin = checkint(begin)
    if begin <= 0 then
        begin = #dest + 1
    end

    local len = #src
    for i = 0, len - 1 do
        dest[i + begin] = src[i + 1]
    end
end
function table.indexof(array, value, begin)
    for i = begin or 1, #array do
        if array[i] == value then
            return i
        end
    end
    return false
end
function table.keyof(hashtable, value)
    for k, v in pairs(hashtable) do
        if v == value then
            return k
        end
    end
    return nil
end
function table.removebyvalue(array, value, removeall)
    local c, i, max = 0, 1, #array
    while i <= max do
        if array[i] == value then
            table.remove(array, i)
            c = c + 1
            i = i - 1
            max = max - 1
            if not removeall then
                break
            end
        end
        i = i + 1
    end
    return c
end
function table.map(t, fn)
    for k, v in pairs(t) do
        t[k] = fn(v, k)
    end
end
function table.walk(t, fn)
    for k, v in pairs(t) do
        fn(v, k)
    end
end
function table.filter(t, fn)
    for k, v in pairs(t) do
        if not fn(v, k) then
            t[k] = nil
        end
    end
end
function table.unique(t, bArray)
    local check = {}
    local n = {}
    local idx = 1
    for k, v in pairs(t) do
        if not check[v] then
            if bArray then
                n[idx] = v
                idx = idx + 1
            else
                n[k] = v
            end
            check[v] = true
        end
    end
    return n
end